# Copyright © 2022 Arrikto Inc.  All Rights Reserved.

"""Kale SDK.

This script uses an ML pipeline to train and serve an Tensorflow Model.
"""

import os

import numpy as np
import tensorflow as tf

from pathlib import Path
from typing import Tuple

from tensorflow.keras import layers

from kale.types import MarshalData
from kale.sdk import pipeline, step


def _split_data(images, labels, train_size=0.9, shuffle=False):
    # get the total size of the dataset
    size = len(images)
    # make an indices array and shuffle it, if required
    indices = np.arange(size)
    if shuffle:
        np.random.shuffle(indices)
    # get the size of training samples
    train_samples = int(size * train_size)
    # split data into training and validation sets
    x_train = images[indices[:train_samples]]
    y_train = labels[indices[:train_samples]]
    x_valid = images[indices[train_samples:]]
    y_valid = labels[indices[train_samples:]]
    return x_train, x_valid, y_train, y_valid


def _get_preprocessed_datasets(x_train, y_train, x_valid, y_valid,
                               batch_size, characters):

    def _encode_single_sample(img_path, label, img_width=200, img_height=50):
        # read image
        img = tf.io.read_file(img_path)
        # decode and convert to grayscale
        img = tf.io.decode_png(img, channels=1)
        # convert to float32 in [0, 1] range
        img = tf.image.convert_image_dtype(img, tf.float32)
        # resize to the desired size
        img = tf.image.resize(img, [img_height, img_width])
        # transpose the image because we want the time
        # dimension to correspond to the width of the image.
        img = tf.transpose(img, perm=[1, 0, 2])
        # map the characters in label to numbers
        label = char_to_num(
            tf.strings.unicode_split(label, input_encoding="UTF-8"))
        # return a dict as our model is expecting two inputs
        return {"image": img, "label": label}

    # mapping characters to integers
    char_to_num = layers.StringLookup(vocabulary=list(characters),
                                      mask_token=None)

    # Training dataset object
    train_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train))
    train_dataset = (train_dataset.map(_encode_single_sample, tf.data.AUTOTUNE)
                                  .batch(batch_size)
                                  .prefetch(buffer_size=tf.data.AUTOTUNE))

    # Validation dataset object
    valid_dataset = tf.data.Dataset.from_tensor_slices((x_valid, y_valid))
    valid_dataset = (valid_dataset.map(_encode_single_sample, tf.data.AUTOTUNE)
                                  .batch(batch_size)
                                  .prefetch(buffer_size=tf.data.AUTOTUNE))
    return train_dataset, valid_dataset


class CTCLayer(layers.Layer):
    """CTC loss layer."""

    def __init__(self, name=None):
        super().__init__(name=name)
        self.loss_fn = tf.keras.backend.ctc_batch_cost

    def call(self, y_true, y_pred):
        # compute the training-time loss value and add it
        # to the layer using `self.add_loss()`.
        batch_len = tf.cast(tf.shape(y_true)[0], dtype="int64")
        input_length = tf.cast(tf.shape(y_pred)[1], dtype="int64")
        label_length = tf.cast(tf.shape(y_true)[1], dtype="int64")

        input_length = input_length * tf.ones(shape=(batch_len, 1),
                                              dtype="int64")
        label_length = label_length * tf.ones(shape=(batch_len, 1),
                                              dtype="int64")

        loss = self.loss_fn(y_true, y_pred, input_length, label_length)
        self.add_loss(loss)

        # at test time, just return the computed predictions
        return y_pred


@step(name="data_loading")
def load_split_dataset() -> Tuple[MarshalData, MarshalData, MarshalData,
                                  MarshalData, MarshalData]:
    """Load and split dataset."""
    # path to the data directory
    data_dir = Path("./captcha_images_v2/")

    # get list of all the images
    images = sorted(list(map(str, list(data_dir.glob("*.png")))))
    labels = [img.split(os.path.sep)[-1].split(".png")[0] for img in images]
    characters = sorted(list(set(char for label in labels for char in label)))

    # Splitting data into training and validation sets
    x_train, x_valid, y_train, y_valid = _split_data(np.array(images),
                                                     np.array(labels))
    return x_train, x_valid, y_train, y_valid, characters


@step(name='model_definition')
def build_model(img_width: int, img_height: int,
                characters: MarshalData) -> MarshalData:
    # Mapping characters to integers
    char_to_num = layers.StringLookup(vocabulary=list(characters),
                                      mask_token=None)
    # Inputs to the model
    inputs = layers.Input(shape=(img_width, img_height, 1), name="image",
                          dtype="float32")

    labels = layers.Input(name="label", shape=(None,), dtype="float32")

    # First conv block
    x = layers.Conv2D(32, (3, 3), activation="relu",
                      kernel_initializer="he_normal",
                      padding="same", name="Conv1")(inputs)
    x = layers.MaxPooling2D((2, 2), name="pool1")(x)

    # Second conv block
    x = layers.Conv2D(64, (3, 3), activation="relu",
                      kernel_initializer="he_normal",
                      padding="same", name="Conv2")(x)
    x = layers.MaxPooling2D((2, 2), name="pool2")(x)

    # We have used two max pool with pool size and strides 2.
    # Hence, downsampled feature maps are 4x smaller. The number of
    # filters in the last layer is 64. Reshape accordingly before
    # passing the output to the RNN part of the model
    new_shape = ((img_width // 4), (img_height // 4) * 64)
    x = layers.Reshape(target_shape=new_shape, name="reshape")(x)
    x = layers.Dense(64, activation="relu", name="dense1")(x)
    x = layers.Dropout(0.2)(x)

    # RNNs
    x = layers.Bidirectional(
        layers.LSTM(128, return_sequences=True, dropout=0.25))(x)
    x = layers.Bidirectional(
        layers.LSTM(64, return_sequences=True, dropout=0.25))(x)

    # Output layer
    x = layers.Dense(len(char_to_num.get_vocabulary()) + 1,
                     activation="softmax",
                     name="dense2")(x)

    # Add CTC layer for calculating CTC loss at each step
    output = CTCLayer(name="ctc_loss")(labels, x)

    # Define the model
    model = tf.keras.Model(inputs=[inputs, labels],
                           outputs=output,
                           name="ocr_model_v1")
    return model


@step(name="model_training")
def train(model: MarshalData, x_train: MarshalData, y_train: MarshalData,
          x_valid: MarshalData, y_valid: MarshalData,
          batch_size: int, characters: MarshalData, epochs: int,
          early_stopping_patience: int) -> MarshalData:
    """Train a Tensorflow model."""
    # Get datasets
    train_dataset, validation_dataset = _get_preprocessed_datasets(x_train,
                                                                   y_train,
                                                                   x_valid,
                                                                   y_valid,
                                                                   batch_size,
                                                                   characters)
    # Add early stopping
    early_stopping = tf.keras.callbacks.EarlyStopping(
        monitor="val_loss", patience=early_stopping_patience,
        restore_best_weights=True)

    # Compile the model and return
    model.compile(optimizer="adam")

    # Train the model
    model.fit(train_dataset, validation_data=validation_dataset,
              epochs=epochs, callbacks=[early_stopping])

    return model


@pipeline(name="tensorflow", experiment="tensorflow-tutorial")
def ml_pipeline(img_width: int = 200, img_height: int = 50,
                batch_size: int = 16, epochs: int = 100,
                early_stopping_patience: int = 5):
    """Run the ML pipeline."""
    x_train, x_valid, y_train, y_valid, characters = load_split_dataset()
    model = build_model(img_width, img_height, characters)
    train(model, x_train, y_train, x_valid, y_valid,
          batch_size, characters, epochs, early_stopping_patience)


if __name__ == "__main__":
    ml_pipeline()
